#include <iostream>
#include <sstream>
#include <cassert>

#if defined( _MSC_VER )
    // This include file has Microsoft specific functions for monitoring dynamic memory usage
#   include <crtdbg.h>
#else
#   define _CrtMemState int
#   define _CrtMemCheckpoint( x )
#   define _CrtMemDifference( x, y, z ) 0
#endif

#include "po.h"         // Include the header file generated by LMX
#include "lmxparse.h"   // Include parser header file.  Only required where marshaling and unmarshaling is done.


// This is extracted from XML Schema Part 0.  The " characters have been modified to ' for easier 
// -reading when embedded in a C++ string.
const char po_xml[] = 
    "<?xml version='1.0'?>\n"
    "<purchaseOrder orderDate='1999-10-20'>\n"
    "    <shipTo country='US'>\n"
    "        <name>Alice Smith</name>\n"
    "        <street>123 Maple Street</street>\n"
    "        <city>Mill Valley</city>\n"
    "        <state>CA</state>\n"
    "        <zip>90952</zip>\n"
    "    </shipTo>\n"
    "    <billTo country='US'>\n"
    "        <name>Robert Smith</name>\n"
    "        <street>8 Oak Avenue</street>\n"
    "        <city>Old Town</city>\n"
    "        <state>PA</state>\n"
    "        <zip>95819</zip>\n"
    "    </billTo>\n"
    "    <comment>Hurry, my lawn is going wild!</comment>\n"
    "    <items>\n"
    "        <item partNum='872-AA'>\n"
    "            <productName>Lawnmower</productName>\n"
    "            <quantity>1</quantity>\n"
    "            <USPrice>148.95</USPrice>\n"
    "            <comment>Confirm this is electric</comment>\n"
    "        </item>\n"
    "        <item partNum='926-AA'>\n"
    "            <productName>Baby Monitor</productName>\n"
    "            <quantity>1</quantity>\n"
    "            <USPrice>39.98</USPrice>\n"
    "            <shipDate>1999-05-21</shipDate>\n"
    "        </item>\n"
    "    </items>\n"
    "</purchaseOrder>\n";


void handle_po()
{
    try
        {
        // Read in the po XML
        //-------------------
        
        // Declare somehwere to store any error code
        lmx::elmx_error l_error = lmx::ELMX_OK;
        
        // Construct an instance of c_root from the XML contained in the memory at po_xml
        c_root l_po( po_xml, sizeof( po_xml ) - 1, &l_error );

        if( l_error != lmx::ELMX_OK )
            {
            std::cout << "An error occurred while reading XML\n";
            return;
            }


        // Interrogate the parsed XML
        //---------------------------
        if( l_po.getchosen() == c_root::e_purchaseOrder )
            {
            // Make some intermediate variables so that components are easier to reference
            const c_PurchaseOrderType *lp_pot = &l_po.get_purchaseOrder();

            if( lp_pot->isset_orderDate() )
                {
                std::cout << "Ordered on: " << lp_pot->get_orderDate() << "\n";

                if( lp_pot->get_orderDate().get_year() <= 2000 )
                    std::cout << "This was ordered last century - It is now overdue!\n";
                }

            const c_Items *lp_items = &lp_pot->get_items();
            
            float l_total_cost = 0;

            // Casts used to avoid vc7 warnings about converting size_t to unsigned int
            std::cout << "Number of items = " << static_cast<unsigned int>(lp_items->size_item()) << "\n";

            // Print out some information about each order item
            for( size_t l_i=0; l_i<lp_items->size_item(); ++l_i )
                {
                // Casts used to avoid vc7 warnings about converting size_t to unsigned int
                std::cout << "Item " << static_cast<unsigned int>(l_i + 1) << ": " << 
                            static_cast<unsigned int>( lp_items->get_item( l_i ).get_quantity() ) << " off - " <<
                            lmx::as_ascii( lp_items->get_item( l_i ).get_productName() ) <<
                            " (" << lmx::as_ascii( lp_items->get_item( l_i ).get_partNum() ) << ")";

                // The comment is optional, so only print it out if it is present
                if( lp_items->get_item( l_i ).isset_comment() )
                    std::cout << " [" << lmx::as_ascii( lp_items->get_item( l_i ).get_comment() ) << "]";

                std::cout << " - $" << lp_items->get_item( l_i ).get_USPrice() << "\n";
                
                l_total_cost += lp_items->get_item( l_i ).get_USPrice().get_scaled( 2 );
                }

            std::cout << "Total Cost: $" << l_total_cost/100 << "\n";

            std::cout << "Ship to: " << lmx::as_ascii( lp_pot->get_shipTo().get_name() ) << "\n";
            std::cout << "Bill to: " << lmx::as_ascii( lp_pot->get_billTo().get_name() ) << "\n";
            std::cout << "\n";


            // Modify the XML
            //---------------
            c_root l_alt_po( l_po );    // We don't actually need to create a new instance before modifying

            l_alt_po.get_purchaseOrder().get_items().append_item();
            c_Items::c_item *lp_item = &l_alt_po.get_purchaseOrder().get_items().back_item();

            lp_item->set_productName( L"Fence" );
            lp_item->set_quantity( 2 );
            lp_item->set_partNum( L"100-AB" );
            lp_item->set_USPrice( lmx::c_decimal( 12.95, 2U ) );

            // We can also do:
            // lp_item->set_USPrice( 12.95 );    // Although the schema needs to specify the fraction digits
            //                                   // facet in order to get the number of decimal places correct!
            // lp_item->set_USPrice( lmx::c_decimal( 1295, 2 ) );    // e.g. 1295 / (10^2)
            // lp_item->set_USPrice( "12.95" );

            lp_item->set_comment( L"Will this stop the baby getting on the lawn?" );
            
            assert( lp_item->is_occurs_ok() );    // Check sufficient elements and attributes added


            // Write the modified version of the XML
            //--------------------------------------
            std::cout << "\n\nModified XML output\n";

            if( l_alt_po.marshal( std::cout ) == lmx::ELMX_OK )
                std::cout << "Modified XML written successfully\n";
            else
                std::cout << "Error writing Modified XML\n";

            // Also marshal to a file
            if( l_alt_po.marshal( "po-out.xml" ) == lmx::ELMX_OK )
                std::cout << "Modified XML written successfully to file\n";
            else
                std::cout << "Error writing Modified XML to file\n";
            }


        // Write out the original XML
        //---------------------------
        std::cout << "\n\nXML output\n";

        if( l_po.marshal( std::cout ) == lmx::ELMX_OK )
            std::cout << "Original XML written successfully\n";
        else
            std::cout << "Error writing Original XML\n";
            
        }   // Ends of try block
        
    catch( std::bad_alloc & )
        {
        // This is the most likely exception unless you have selected LMX to throw exceptions.
        std::cout << "An out of memory exception has occurred XML\n";
        }
    catch( ... )
        {
        // Other exceptions might include iostream exceptions etc.
        std::cout << "Another exception has occurred\n";
        }
}


int main( int argc, char **argv )
{
    std::cout << "LMX Example Program\n";

    {
    // The iostream functions dynamically allocate memory the first time they 
    // -are called.  They are not cleared until the program ends.  So that 
    // -these allocations do not affect the heap checking stats, dummy uses 
    // -of the libraries are made so that they are initialised.  We can then 
    // -checkpoint the heap after this point.
    std::ostringstream l_t1;
    l_t1 << "" << 12 << false;
    std::ofstream l_outf( "po-out.xml" );
    #if defined( LMX_DEBUG_CAPTURE_ERROR ) && LMX_DEBUG_CAPTURE_ERROR == 1
        #if defined( _MSC_VER ) && _MSC_VER >= 1300
            lmx::debug_error.item_name.reserve( 100 );  // For VC7.1 & 8
        #endif
    #endif
    }


    _CrtMemState s1, s2, s3;
    // Store a memory checkpoint in the s1 memory-state structure
    _CrtMemCheckpoint( &s1 );


    // The main part of the example happens in this function
    handle_po();


    _CrtMemCheckpoint( &s2 );
    if( ! _CrtMemDifference( &s3, &s1, &s2 ) )
        std::cout << "Heap OK\n";
    else
        std::cout << "The heap has leaked!!!\n";


    return 0;
}